import torch
import numpy as np

from .IVFPQBase import IVFPQBase
from .IVFPQTopk import IVFPQTopk
from .PQ import PQ
from .MPQ import MPQ
from .kmeans import KMeans

class IVFMPQ:
  def __init__(self):
    pass
      